--
--
-- Better AttractChase by TehRealSalt
--

local function ReplaceMe(mo, newType)
	mo.type = newType;

	mo.radius = FixedMul(mo.info.radius, mo.scale);
	mo.height = FixedMul(mo.info.height, mo.scale);
	mo.flags = mo.info.flags;

	if (mo.info.spawnhealth)
		mo.health = mo.info.spawnhealth;
	else
		mo.health = 1;
	end

	mo.reactiontime = mo.info.reactiontime;
	mo.state = mo.info.spawnstate;
end

local function AttractMaxDist(player)
	local attract = 0; -- 96

	if (player.powers[pw_shield] & SH_PROTECTELECTRIC)
		attract = $1 + 768;
	end

	return attract;
end

local DEFAULT_ATTRACT_ACCEL = (9 << FRACBITS);
local function Attract(source, dest, accel)
	if not (source and source.valid)
	or (source.health <= 0)
		return;
	end

	if not (dest and dest.valid)
	or (dest.health <= 0)
		return;
	end

	local sPos = Vector3.New(
		source.x, source.y,
		source.z + (source.height / 2)
	);
	local dPos = Vector3.New(
		dest.x, dest.y,
		dest.z + (dest.height / 2)
	);

	if (accel == nil)
		accel = DEFAULT_ATTRACT_ACCEL;
	end

	local MAX_RING_ACCEL = FixedMul(accel, source.scale);
	local MIN_RING_ACCEL = MAX_RING_ACCEL >> 2;

	local dir = Vector3.DirectionTo(sPos, dPos);
	local mom = Vector3.MobjMomentumVector(source);

	-- Fluid
	--[[
	local dot = Vector3.Dot(Vector3.Normalize(mom), dir);

	local mul = ((dot + FRACUNIT) / 2); -- Range from 0 to FRACUNIT
	local accel = MIN_RING_ACCEL + FixedMul(MAX_RING_ACCEL - MIN_RING_ACCEL, FRACUNIT - mul);

	Vector3.SetMobjMomentum(source, mom + Vector3.Mul(accel, dir));
	--]]

	-- Rigid
	local accel = {
		[-1] = MAX_RING_ACCEL,
		[0] = MAX_RING_ACCEL,
		[1] = MIN_RING_ACCEL,
	};
	local sd = Vector3.IntegerSign(dir);
	local sm = Vector3.IntegerSign(mom);

	Vector3.SetMobjMomentum(
		source,
		mom + (Vector3.Mul(Vector3.New(
			accel[sm.x * sd.x],
			accel[sm.y * sd.y],
			accel[sm.z * sd.z]
		), dir))
	);
end

local function NewAttractChase(actor, accel, doReplace)
	if ((actor.flags & MF_NOCLIPTHING) or (actor.flags2 & MF2_NIGHTSPULL) or not actor.health)
		return;
	end

	-- Flicker before disappearing.
	if ((leveltime & 1) and actor.fuse > 0 and actor.fuse < 2*TICRATE)
		actor.flags2 = $1 | MF2_DONTDRAW;
	else
		actor.flags2 = $1 & ~MF2_DONTDRAW;
	end

	-- Turn fling rings back into regular rings if attracted.
	--[[
	if (actor.tracer and actor.tracer.valid)
	and (actor.tracer.player and actor.tracer.player.valid)
	and (actor.info.reactiontime and actor.type != actor.info.reactiontime)
	and not (player.powers[pw_shield] & SH_PROTECTELECTRIC)
		ReplaceMe(actor, actor.info.reactiontime);
		return;
	end
	--]]

	if not (actor.tracer and actor.tracer.valid)
	or not (actor.tracer.player and actor.tracer.player.valid)
	or not (actor.tracer.health)
	--or not (P_CheckSight(actor, actor.tracer))
		if (actor.ringAttracted)
			actor.flags = ($1 & ~(MF_NOCLIP|MF_NOCLIPHEIGHT)) | (actor.info.flags & (MF_NOCLIP|MF_NOCLIPHEIGHT));
			actor.tracer = nil;

			if not (P_CheckPosition(actor, actor.x, actor.y))
				-- Stuck in the void?
				P_RemoveMobj(actor);
				return;
			end

			actor.ringAttracted = nil;
		end

		return;
	end

	-- If a fling ring gets attracted, change it into a normal ring.
	if (doReplace) and (actor.type == actor.info.reactiontime)
		ReplaceMe(actor, actor.info.painchance);
		return;
	end

	-- Let attracted rings move through walls and such.
	actor.flags = $1 | (MF_NOCLIP|MF_NOCLIPHEIGHT);
	actor.ringAttracted = true;

	Attract(actor, actor.tracer, accel);
end

function A_AttractChase(actor)
	actor.usingAttractChase = true;
	NewAttractChase(actor, nil, nil, true);
end

addHook("PlayerThink", function(player)
	if not (player and player.valid and player.mo and player.mo.valid and player.mo.health)
		return;
	end

	local ourMaxDist = AttractMaxDist(player);
	if (ourMaxDist <= 0)
		-- Can't attract
		return;
	end

	ourMaxDist = FixedMul($1 << FRACBITS, player.mo.scale) >> FRACBITS;

	local function LookForRings(_, mobj)
		if not (mobj.usingAttractChase)
			-- Not an object using A_AttractChase
			return nil;
		end

		if (mobj.health <= 0 or mobj.threshold > 0)
			return nil;
		end

		if (mobj.target and mobj.target.valid) and (player.mo == mobj.target)
			-- Fired as projectile, don't home in on the owner.
			return nil;
		end

		if (mobj.type == MT_REDTEAMRING and player.ctfteam != 1)
		or (mobj.type == MT_BLUETEAMRING and player.ctfteam != 2)
			-- When in CTF, don't pull rings that you cannot pick up.
			return nil;
		end

		local xdist = (mobj.x / FRACUNIT) - (player.mo.x / FRACUNIT);
		local ydist = (mobj.y / FRACUNIT) - (player.mo.y / FRACUNIT);
		local zdist = ((mobj.z + (mobj.height / 2)) / FRACUNIT) - ((player.mo.z + (player.mo.height / 2)) / FRACUNIT);

		local dist = P_AproxDistance(P_AproxDistance(xdist, ydist), zdist);

		if (dist < ourMaxDist)
			mobj.tracer = player.mo;

			if (mobj.hnext and mobj.hnext.valid)
				mobj.hnext.hprev = mobj.hprev;
			end

			if (mobj.hprev and mobj.hprev.valid)
				mobj.hprev.hnext = mobj.hnext;
			end
		end

		return nil;
	end

	// Just to be safe, add a little bit past
	// the block that we would search.
	local MAXRADIUS = 64 << FRACBITS;

	searchBlockmap("objects", LookForRings, player.mo,
		player.mo.x - (ourMaxDist << FRACBITS) - MAXRADIUS, player.mo.x + (ourMaxDist << FRACBITS) + MAXRADIUS,
		player.mo.y - (ourMaxDist << FRACBITS) - MAXRADIUS, player.mo.y + (ourMaxDist << FRACBITS) + MAXRADIUS
	);
end);
